/* simple bytea and base64 encoders/decoders              */
/* (C) 2001 by Joerg Hessdoerfer                          */
/* Questions and comments: Joerg.Hessdoerfer@sea-gmbh.com */

#include <stdio.h>
#include <stdlib.h>

#include "bytea_codes.h"

/* return the size of bytea encoded memory block */
size_t bytea_encoded_size(const unsigned char *src, size_t size)
{
  size_t i,sz;
  const unsigned char *s;

  for(s = src, i = 0, sz = 0; i < size; i++, s++) sz += bytea_codelen[*s];

  return sz;
}

/* return the size of decoded bytea memory block */
size_t bytea_decoded_size(const unsigned char *src, size_t size)
{
  size_t i,sz;
  const unsigned char *s;

  for(s = src, i = 0, sz = 0; i < size; sz++)
    {
      if(*s++ == '\\')
	{
	  i += 4;
	  
	  if(*s++ == '\\')
	    {
	      i += 1; /* 0 is double escaped!! */
	      s++;
	    }
	  s+=2;
	}
      else i++;
    }
      
  return sz;
}

/* this routine REQUIRES a pre-allocated memory block for encoding */
/* result. Please use bytea_encoded_size() to calculate the required */
/* mem size */
void encode_bytea_prealloced(const unsigned char *src, unsigned char *dst, size_t size)
{
  size_t i,j;
  const unsigned char *s;
  unsigned char *d;

  s = src;
  d = dst;
  
  for(i = 0; i < size; i++, s++)
    {
      for(j = 0; j < bytea_codelen[*s]; j++) *d++ = bytea_codes[*s][j];
    }
}

/* this routine REQUIRES a pre-allocated memory block for encoding */
/* result. Please use bytea_decoded_size() to calculate the required */
/* mem size */
void decode_bytea_prealloced(const unsigned char *src, unsigned char *dst, size_t size)
{
  size_t i;
  const unsigned char *s;
  unsigned char *d;

  s = src;
  d = dst;
  
  for(i = 0; i < size;)
    {
      if(*s == '\\')
	{
	  s++;
	  
	  i += 4;
	  
	  if(*s == '\\')
	    {
	      i += 1; /* 0 is double escaped!! */
	      s += 4;
	      *d++ = 0;
	    }
	  else
	    {
	      *d++ =
		((*s - '0') << 6) +
		((*(s+1) - '0') << 3) +
		(*(s+2) - '0');
	      s += 3;
	    }
	  
	}
      else
	{
	  *d++ = *s++;
	  i++;
	}
    }
      
}


/* this routine returns a pointer to a block of bytea-encoded memory, with a trailing 0 */
/* returned ptr must be free()ed */
unsigned char *encode_bytea(const unsigned char *src, size_t size, size_t *used)
{
  size_t sz;
  unsigned char *blk;
  
  sz = bytea_encoded_size(src, size);

  blk = (unsigned char *)malloc(sz + 1);
  if(blk == NULL) return NULL;

  encode_bytea_prealloced(src, blk, size);
  blk[sz] = 0;
  
  *used = sz;
  
  return blk;
}


/* this routine returns a pointer to a block of memory decoded from bytea input */
/* warning: no trailing 0! */
/* returned ptr must be free()ed */
unsigned char *decode_bytea(const unsigned char *src, size_t size, size_t *used)
{
  size_t sz;
  unsigned char *blk;
  
  sz = bytea_decoded_size(src, size);

  blk = (unsigned char *)malloc(sz);
  if(blk == NULL) return NULL;

  decode_bytea_prealloced(src, blk, size);

  *used = sz;
  
  return blk;
}

/* ==================== base 64 section ======================================= */

/* return the size of base64 encoded memory block */
size_t base64_encoded_size(const unsigned char *src, size_t size, int with_lf)
{
  size_t sz;
  
  sz = ((size + 2) / 3) * 4;
  if(with_lf) sz += (sz / 72); /* add space for <LF> */
  
  return sz;
}

/* return the size of decoded base64 memory block */
size_t base64_decoded_size(const unsigned char *src, size_t size)
{
  size_t i, sz, r;
  const unsigned char *s;

  /* we have to loop through here, to filter out unwanted chars! */
  for(s = src, i = 0, sz = 0; i < size; i++, s++)
    {
      if(base64_dec_table[*s] < 64) sz++;
    }

  r = sz - ((sz / 4)*4);
  sz = (sz / 4) * 3;

  /* r == 1 (meaning 3 trailing '=') is impossible! */
  sz += (r == 0 ? 0 : (r == 2 ? 1 : 2));
  
  return sz;
}

/* this routine REQUIRES a pre-allocated memory block for encoding */
/* result. Please use base64_encoded_size() to calculate the required */
/* mem size */
void encode_base64_prealloced(const unsigned char *src, unsigned char *dst, size_t size, int with_lf)
{
  size_t i,j,no_lf;
  const unsigned char *s;
  unsigned char *d;
  long l;


  s = src;
  d = dst;

  for(no_lf = 0, i = 0; i < size/3; i++, s+=3)
    {
      l = (*s << 16) + (*(s+1) << 8) + *(s+2);
      *d++ = base64_enc_table[(l >> 18) & 0x3f];
      *d++ = base64_enc_table[(l >> 12) & 0x3f];
      *d++ = base64_enc_table[(l >>  6) & 0x3f];
      *d++ = base64_enc_table[ l        & 0x3f];

      if(with_lf)
	{
	  no_lf += 4;
	  if(no_lf >= 72)
	    {
	      *d++ = '\n';
	      no_lf = 0;
	    }
	}
    }
  if(size - (size/3)*3 == 2)
    {
      l = (*s << 8) + *(s+1);
      
      *d++ = base64_enc_table[(l >> 12) & 0x3f];
      *d++ = base64_enc_table[(l >>  6) & 0x3f];
      *d++ = base64_enc_table[ l        & 0x3f];
      *d++ = '=';
    }
  else if(size - (size/3)*3 == 1)
    {
      l = *s;

      *d++ = base64_enc_table[(l >>  6) & 0x3f];
      *d++ = base64_enc_table[ l        & 0x3f];
      *d++ = '=';
      *d++ = '=';
    }

}

/* this routine REQUIRES a pre-allocated memory block for decoding */
/* result. Please use base64_decoded_size() to calculate the required */
/* mem size */
void decode_base64_prealloced(const unsigned char *src, unsigned char *dst, size_t size)
{
  size_t i;
  const unsigned char *s;
  unsigned char *d;
  int j;

  s = src;
  d = dst;
  
  for(i = 0; i < size; )
    {
      long l;

      for (l = 0, j = 0; j < 4 && i < size; i++, s++)
	{
	  if(base64_dec_table[*s] < 64)
	    {
	      l += base64_dec_table[*s] << (3-j)*6;
	      j++;
	    }
	}

      if(j == 4)
	{
	  *d++ = (l >> 16) & 0xff;
	  *d++ = (l >>  8) & 0xff;
	  *d++ =  l        & 0xff;
	}
      else if(j == 3)
	{
	  *d++ = (l >> 14) & 0xff;
	  *d++ = (l >>  6) & 0xff;
	}
      else if(j == 2)
	{
	  *d++ = (l >> 12) & 0xff;
	}
    }
}


/* this routine returns a pointer to a block of base64-encoded memory, with a trailing 0 */
/* returned ptr must be free()ed */
unsigned char *encode_base64(const unsigned char *src, size_t size, size_t *used, int with_lf)
{
  size_t sz;
  unsigned char *blk;
  
  sz = base64_encoded_size(src, size, with_lf);

  blk = (unsigned char *)malloc(sz + 1);
  if(blk == NULL) return NULL;

  encode_base64_prealloced(src, blk, size, with_lf);
  blk[sz] = 0;

  *used = sz;
  
  return blk;
}


/* this routine returns a pointer to a block of memory decoded from base64 input */
/* warning: no trailing 0! */
/* returned ptr must be free()ed */
unsigned char *decode_base64(const unsigned char *src, size_t size, size_t *used)
{
  size_t sz;
  unsigned char *blk;
  
  sz = base64_decoded_size(src, size);

  blk = (unsigned char *)malloc(sz);
  if(blk == NULL) return NULL;

  decode_base64_prealloced(src, blk, size);

  *used = sz;
  
  return blk;
}
